import random

from .utils import ExecuteCommand, ExecuteCommandWithOutput


def CheckForwarding() -> None:
    cmd = "cat /proc/sys/net/ipv4/ip_forward"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv4.ip_forward=1\" >> /etc/sysctl.conf'")

    cmd = "cat /proc/sys/net/ipv4/conf/all/forwarding"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv4.conf.all.forwarding=1\" >> /etc/sysctl.conf'")

    cmd = "cat /proc/sys/net/ipv6/conf/all/forwarding"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv6.conf.all.forwarding=1\" >> /etc/sysctl.conf'")

    ExecuteCommand("sudo sysctl -p")


def GetMacAddress() -> str:
    mac = "52:54:00:%02x:%02x:%02x" % (
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255))

    return mac


def AddLinuxBridge(name: str, ip4: str, ip6: str, plen4: int, plen6: int) -> None:
    mac_address = GetMacAddress()
    dummy_intf = "{}-nic".format(name)
    ExecuteCommand("sudo ip link add {} address {} type dummy".format(dummy_intf, mac_address))
    ExecuteCommand("sudo brctl addbr {}".format(name))
    ExecuteCommand("sudo brctl stp {} on".format(name))
    ExecuteCommand("sudo brctl addif {} {}".format(name, dummy_intf))
    ExecuteCommand("sudo ip address add {}/{} dev {}".format(ip4, plen4, name))
    if ip6 != "None":
        ExecuteCommand("sudo ip -6 address add {}/{} dev {}".format(ip6, plen6, name))
    ExecuteCommand("sudo ip link set dev {} up".format(dummy_intf))
    ExecuteCommand("sudo ip link set dev {} up".format(name))


def DelLinuxBridge(name):
    dummy_intf = f"{name}-nic"
    ExecuteCommand(f"sudo ip link set dev {dummy_intf} down")
    ExecuteCommand(f"sudo ip link set dev {name} down")
    ExecuteCommand(f"sudo ip link del {dummy_intf}")
    ExecuteCommand(f"sudo ip link del {name}")


def AddDelIptableRules(op, name, nw4):
    ExecuteCommand(f"sudo iptables -t mangle -{op} POSTROUTING -o {name} -p udp "
                   "-m udp --dport 68 -j CHECKSUM --checksum-fill")
    ExecuteCommand(f"sudo iptables -t nat -{op} POSTROUTING -s {nw4} -d 224.0.0.0/24 -j RETURN")
    ExecuteCommand(f"sudo iptables -t nat -{op} POSTROUTING -s {nw4} -d "
                   "255.255.255.255/32 -j RETURN")
    ExecuteCommand(f"sudo iptables -t nat -{op} POSTROUTING -s {nw4} ! -d {nw4} "
                   "-p tcp -j MASQUERADE --to-ports 1024-65535")
    ExecuteCommand(f"sudo iptables -t nat -{op} POSTROUTING -s {nw4} ! -d {nw4} "
                   "-p udp -j MASQUERADE --to-ports 1024-65535")
    ExecuteCommand(f"sudo iptables -t nat -{op} POSTROUTING -s {nw4} ! -d {nw4} -j MASQUERADE")
    ExecuteCommand(f"sudo iptables -t filter -{op} INPUT -i {name} -p udp -m udp "
                   "-m multiport --dports 53,67 -j ACCEPT")
    ExecuteCommand(f"sudo iptables -t filter -{op} INPUT -i {name} -p tcp -m tcp "
                   "-m multiport --dports 53,67 -j ACCEPT")
    ExecuteCommand(f"sudo iptables -t filter -{op} FORWARD -d {nw4} -o {name} -j ACCEPT")
    ExecuteCommand(f"sudo iptables -t filter -{op} FORWARD -s {nw4} -i {name} -j ACCEPT")
    ExecuteCommand(f"sudo iptables -t filter -{op} FORWARD -i {name} -j ACCEPT")
    ExecuteCommand(f"sudo iptables -t filter -{op} FORWARD -o {name} -j ACCEPT")
    ExecuteCommand("sudo sh -c 'iptables-save > /etc/iptables/rules.v4'")



from pathlib import Path

OP_CREATE = 1
OP_DELETE = 2
OP = OP_CREATE

DRY_RUN = False
NO_NETWORK = False
NO_VM = False

NETWORK_TYPES = ["nat", "isolated", "management"]
VM_FLAVORS = ["pe", "ce", "dev"]

LIBVIRT_BASE_PATH = Path('/var/lib/libvirt')
LIBVIRT_IMAGES = LIBVIRT_BASE_PATH.joinpath('images')
LIBVIRT_TEMPLATES = LIBVIRT_IMAGES.joinpath('templates')
LIBVIRT_QEMU_BASE = Path('/etc/libvirt/qemu')
LIBVIRT_QEMU_NW = LIBVIRT_QEMU_BASE.joinpath('networks')

BASE_OS = "ubuntu"
UBUNTU_TEMPLATE = LIBVIRT_TEMPLATES.joinpath('ubuntu-22.04-server.qcow2')
ROCKY_TEMPLATE = LIBVIRT_TEMPLATES.joinpath('rocky-8.qcow2')
OS_IMAGE_TEMPLATE = UBUNTU_TEMPLATE

NAT_NW_BASE = Path.home().joinpath('.topology/bridges')



import logging
import sys

import deployer.globals as G
from deployer.network import Network
from deployer.topology import Topology
from deployer.utils import ProcessArguments
from deployer.virtual_machine import VirtualMachine


class Deployer:
    config_ = {}
    topology_ = Topology()

    def __init__(self, config):
        self.config_ = config
    # end __init__

    def _is_valid_version(self):
        if self.config_ is None:
            logging.error(f"Empty config : {self.config_}")
            return False

        if self.config_.get("version", None) is None:
            logging.error("This looks like older version of config which is "
                          " supported only by 0.0.x version")
            return False

        if self.config_["version"] != 2:
            logging.error("This version supports only version 2 of config")
            return False

        return True
    # end _is_valid_version

    def _parse_networks(self):
        nw_configs = self.config_.get("networks", None)

        if nw_configs is None:
            logging.error("No network config found. Exiting")
            return False

        if not isinstance(nw_configs, list):
            logging.error("'network' config should be a list")
            return False

        if len(nw_configs) == 0:
            logging.error("'network' config cannot be empty list")
            return False

        for nw_conf in nw_configs:
            nw = Network(nw_conf)
            if nw is None:
                return False
            name = nw.name_
            self.Topology().AddNetwork(name, nw)

        return True
    # end _parse_networks

    def _parse_vms(self):
        vm_configs = self.config_.get("vms", None)

        if vm_configs is None:
            logging.error("No VMs config found. Exiting")
            return False

        if not isinstance(vm_configs, list):
            logging.error("'vms' config should be a list")
            return False

        if len(vm_configs) == 0:
            logging.error("'vms' config cannot be empty list")
            return False

        for vm_conf in vm_configs:
            vm = VirtualMachine(vm_conf)
            if vm is None:
                return False
            name = vm.name_
            self.Topology().AddVm(name, vm)

        return True
    # end _parse_vms

    def ParseConfig(self):
        if not self._is_valid_version():
            return False

        if not self._parse_networks():
            return False

        if not self._parse_vms():
            return False

        return True
    # end ParseConfig

    def Topology(self):
        return self.topology_
    # end Topology

# end Deployer


def topology_deployer():
    config = ProcessArguments(sys.argv[1:])
    tDeployer = Deployer(config)
    if not tDeployer.ParseConfig():
        sys.exit(1)

    topology = tDeployer.Topology()

    if G.OP == G.OP_CREATE:
        topology.Create()
    elif G.OP == G.OP_DELETE:
        topology.Delete()
    else:
        logging.error(f"Invalid operation {G.OP}")
        sys.exit(1)



import logging
import sys

import deployer.globals as globals


class Topology:
    networks_ = {}
    vms_ = {}

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Topology, cls).__new__(cls)
        return cls.instance

    def AddNetwork(self, name, nw):
        if self.__is_existing_network(name):
            logging.error(f"Network {name} already exists")
            sys.exit(2)

        self.networks_[name] = nw
    # end AddNetwork

    def AddVm(self, name, vm):
        if self.__is_existing_vm(name):
            logging.error(f"Virtual Machine {name} already exists")
            sys.exit(2)

        self.vms_[name] = vm
    # end AddVm

    def Delete(self):
        if not globals.NO_VM:
            self.__delete_vms()

        if not globals.NO_NETWORK:
            self.__delete_networks()
    # end Delete

    def Create(self):
        if not globals.NO_NETWORK:
            self.__create_networks()

        if not globals.NO_VM:
            self.__create_vms()

        # Add Print Topology here
    # end __create

    def GetNetwork(self, name):
        return self.networks_.get(name, None)
    # end GetNetwork

    def Networks(self):
        return self.networks_.items()
    # end Networks

    def Vms(self):
        return self.vms_.items()
    # end Vms

    def __create_networks(self):
        for _, nw in self.networks_.items():
            nw.Create()
    # end __create_networks

    def __create_vms(self):
        for _, vm in self.vms_.items():
            vm.Create()
    # end __create_vms

    def __delete_networks(self):
        for _, nw in self.networks_.items():
            nw.Delete()
    # end __delete_networks

    def __delete_vms(self):
        for _, vm in self.vms_.items():
            vm.Delete()
    # end __delete_vms

    def __is_existing_network(self, name):
        nw = self.networks_.get(name, None)
        if nw is None:
            return False
        return True
    # end __is_existing_network

    def __is_existing_vm(self, name):
        vm = self.vms_.get(name, None)
        if vm is None:
            return False
        return True
    # end __is_existing_vm



import argparse
import json
import logging
import shlex
import subprocess
import sys

import deployer.globals as G
from .globals import OP_CREATE, OP_DELETE, ROCKY_TEMPLATE, UBUNTU_TEMPLATE


def ProcessArguments(cli_args):
    parser = argparse.ArgumentParser(description="Parse topology config")

    parser.add_argument("-c", "--config", required=True,
                        help="JSON Config file that defines the topology")

    parser.add_argument("-o", "--operation", type=str, required=True,
                        choices=["create", "delete", "CREATE", "DELETE"], default="create",
                        help="Operation to create or delete topology from the JSON Config")

    parser.add_argument("-i", "--image", type=str, required=True,
                        choices=["ubuntu", "rocky"], default="ubuntu",
                        help="Choose the base OS image for the VMS")

    parser.add_argument("-l", "--log", type=str, default="ERROR",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                        help="Set the log level")
    parser.add_argument("--dry-run", action="store_true",
                        help="Instead of executing, print the commands")

    skip_operation = parser.add_mutually_exclusive_group()
    skip_operation.add_argument("--skip-network", action="store_true",
                                help="Skip creating networks. Cannot be used with --skip-vm")
    skip_operation.add_argument("--skip-vm", action="store_true",
                                help="Skip creating vms. Cannot be used with --skip-network")

    args = parser.parse_args(cli_args)

    if args.log:
        log_level = getattr(logging, args.log.upper(), None)
        if not isinstance(log_level, int):
            raise ValueError(f"Invalid log level: {args.log}")
        logging.basicConfig(level=log_level)

    if args.dry_run:
        G.DRY_RUN = args.dry_run

    if args.operation.upper() == "CREATE":
        G.OP = OP_CREATE
    else:
        G.OP = OP_DELETE

    if args.skip_network:
        G.NO_NETWORK = args.skip_network

    if args.skip_vm:
        G.NO_VM = args.skip_vm

    if args.image.lower() == "ubuntu":
        G.BASE_OS = "ubuntu"
        G.OS_IMAGE_TEMPLATE = UBUNTU_TEMPLATE
    elif args.image.lower() == "rocky":
        G.BASE_OS = "rocky"
        G.OS_IMAGE_TEMPLATE = ROCKY_TEMPLATE
    else:
        logging.critical("Invalid image argument")
        sys.exit(1)

    with open(args.config) as conf_file:
        config = json.load(conf_file)

    return config


def ExecuteCommand(cmd):
    if not G.DRY_RUN:
        logging.info(f"Executing command : {cmd}")
        ret = subprocess.call(shlex.split(cmd))
        if ret != 0:
            logging.error(f"Command {cmd} failed wih return code : {ret}")
            sys.exit(ret)
    else:
        logging.debug(f"Dry Run enabled. Command : {cmd}")


def ExecuteCommandWithOutput(cmd):
    logging.info(f"Executing command : {cmd}")
    res = subprocess.run(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if res.returncode != 0:
        logging.error(f"Command {cmd} failed wih return code : {res.returncode}")
        logging.error(f"Error message : {res.stderr}")
        sys.exit(res.returncode)
    return res.stdout.decode('utf-8').strip()



import ipaddress
import logging
import psutil
import sys
from xml.dom import minidom
import xml.etree.cElementTree as ET

import deployer.globals as G
from deployer.globals import LIBVIRT_QEMU_NW, NETWORK_TYPES, OP_CREATE
from deployer.nat_utils import AddDelIptableRules, AddLinuxBridge, \
        CheckForwarding, DelLinuxBridge
from deployer.topology import Topology
from deployer.utils import ExecuteCommand


class Network:
    type_ = None
    name_ = ""
    ip4_ = None
    network4_ = None
    network6_ = None
    ip6_ = None
    plen4_ = 24
    plen6_ = 120
    topology_ = Topology()

    def __new__(cls, conf):
        if not Network._validate_network_conf(conf):
            return None
        return super(Network, cls).__new__(cls)
    # end __new__

    def __init__(self, conf):
        self.type_ = conf["type"]
        self.name_ = conf["name"]
        if not self.IsIsolated():
            self._add_v4_network(conf["subnet4"])
            self._add_v6_network(conf.get("subnet6", None))
    # end __init__

    @staticmethod
    def _has_valid_network_fields(nw):
        if nw.get("type", None) not in NETWORK_TYPES:
            logging.error("Invalid network type in network config")
            return False

        if nw.get("name", None) is None:
            logging.error("'name' is required key in network config")
            return False

        if nw["type"] == "isolated":
            return True

        if nw.get("subnet4", None) is None:
            logging.error("'subnet' is required key for nat and management networks")
            return False

        return True
    # end _has_valid_network_fields

    @staticmethod
    def _validate_network_conf(nw):
        if G.NO_NETWORK:
            return True

        if not Network._has_valid_network_fields(nw):
            return False

        if G.OP != OP_CREATE:
            return True

        name = nw["name"]
        ntype = nw["type"]

        if len(name) > 12:
            logging.error(f"Max length for name is 12 characters : {name}")
            return False

        interfaces = psutil.net_if_addrs()

        if name in interfaces:
            logging.error(f"Network with name {name} is already present")
            return False

        if ntype == "isolated":
            return True

        try:
            subnet4 = ipaddress.ip_network(nw["subnet4"],
                                           strict=False)
            subnet6 = None
            if nw.get("subnet6", None) is not None:
                subnet6 = ipaddress.ip_network(nw["subnet6"],
                                               strict=False)
        except ValueError:
            logging.error("Invalid Subnet(s). Please check again")
            return False

        for interface, addrs in psutil.net_if_addrs().items():
            for addr in addrs:
                if addr.family == 2:
                    if ipaddress.IPv4Address(addr.address) in subnet4:
                        logging.error(f"Network colliding with {str(subnet4)} "
                                      f" already exists on {interface}")
                        return False

                if (addr.family == 10 and subnet6 is not None):
                    v6_addr = addr.address.split('%')[0]
                    if ipaddress.IPv6Address(v6_addr) in subnet6:
                        logging.error(f"Network colliding with {str(subnet6)} "
                                      f" already exists on {interface}")
                        return False

        for _, network in Network.Topology().Networks():
            if subnet4 == network.network4_:
                logging.error(f"IPv4 Network {name} is colliding with "
                              f"{network.name_}")
                return False

            if subnet6 is not None:
                if subnet6 == network.network6_:
                    logging.error(f"IPv6 Network {name} is colliding with "
                                  f"{network.name_}")
                    return False

        return True
    # end _validate_network_conf

    @staticmethod
    def Topology():
        return Network.topology_
    # end Topology

    def _add_v4_network(self, subnet4):
        logging.debug(f"Parsing IPv4 Network for {self.name_}")
        self.network4_ = ipaddress.IPv4Network(subnet4)
        self.ip4_ = self.network4_[1]
    # end _add_v4_network

    def _add_v6_network(self, subnet6):
        if subnet6 is not None:
            logging.debug(f"Parsing IPv6 Network for {self.name_}")
            self.network6_ = ipaddress.IPv6Network(subnet6)
            self.ip6_ = self.network6_[1]
        else:
            logging.warning(f"No IPv6 subnet provided for {self.name_}")
    # end _add_v6_network

    def _create_nat_network(self):
        logging.info(f"Creating NAT network : {self.name_}")

        if self.network4_:
            self.plen4_ = self.network4_.prefixlen
        if self.network6_:
            self.plen6_ = self.network6_.prefixlen

        CheckForwarding()

        AddLinuxBridge(self.name_, str(self.ip4_), str(self.ip6_),
                       self.plen4_, self.plen6_)
        AddDelIptableRules("A", self.name_, str(self.network4_))
    # end _create_nat_network

    def _create_management_network(self):
        logging.info(f"Creating Management network : {self.name_}")
        nw_cfg_file = LIBVIRT_QEMU_NW.joinpath(f"{self.name_}.xml")

        nw_ele = ET.Element("network")
        ET.SubElement(nw_ele, "name").text = self.name_
        ET.SubElement(nw_ele, "bridge", {"name": self.name_, "stp": "on", "delay": "0"})

        if self.network4_:
            _ = ET.SubElement(nw_ele, "ip", {
                "address": str(self.ip4_),
                "netmask": str(self.network4_.netmask)
            })

        xmlstr = minidom.parseString(ET.tostring(nw_ele)).toprettyxml(indent="   ")
        with open(nw_cfg_file, "w") as f:
            f.write(xmlstr)

        cmd = f"sudo virsh net-define {nw_cfg_file}"
        ExecuteCommand(cmd)
        cmd = f"sudo virsh net-start {self.name_}"
        ExecuteCommand(cmd)
        cmd = f"sudo virsh net-autostart {self.name_}"
        ExecuteCommand(cmd)
    # end _create_management_network

    def _create_isolated_network(self):
        logging.info(f"Creating Isolated network : {self.name_}")
        nw_cfg_file = LIBVIRT_QEMU_NW.joinpath(f"{self.name_}.xml")

        nw_ele = ET.Element("network")
        ET.SubElement(nw_ele, "name").text = self.name_
        ET.SubElement(nw_ele, "bridge", {"name": self.name_, "stp": "on", "delay": "0"})

        xmlstr = minidom.parseString(ET.tostring(nw_ele)).toprettyxml(indent="   ")
        with open(nw_cfg_file, "w") as f:
            f.write(xmlstr)

        cmd = f"sudo virsh net-define {nw_cfg_file}"
        ExecuteCommand(cmd)
        cmd = f"sudo virsh net-start {self.name_}"
        ExecuteCommand(cmd)
        cmd = f"sudo virsh net-autostart {self.name_}"
        ExecuteCommand(cmd)
    # end _create_isolated_network

    def _delete_nat_network(self):
        AddDelIptableRules("D", self.name_, str(self.network4_))
        DelLinuxBridge(self.name_)
    # end _delete_nat_network

    def _delete_libvirt_network(self):
        cmd = f"sudo virsh net-destroy {self.name_}"
        ExecuteCommand(cmd)
        cmd = f"sudo virsh net-undefine {self.name_}"
        ExecuteCommand(cmd)
    # end _delete_libvirt_network

    def Delete(self):
        if self.IsNat():
            self._delete_nat_network()
        elif self.IsLibvirtNetwork():
            self._delete_libvirt_network()
        else:
            logging.error(f"Unknown network type {self.type_}")
            sys.exit(3)
    # end Delete

    def Create(self):
        if self.IsNat():
            self._create_nat_network()
        elif self.IsIsolated():
            self._create_isolated_network()
        elif self.IsManagement():
            self._create_management_network()
        else:
            logging.error(f"Unknown network type {self.type_}")
            sys.exit(3)
    # end Create

    def Type(self):
        return self.type_
    # end Type

    def IsIsolated(self):
        return self.type_ == "isolated"
    # end IsIsolated

    def IsManagement(self):
        return self.type_ == "management"
    # end IsManagement

    def IsNat(self):
        return self.type_ == "nat"
    # end IsNat

    def IsLibvirtNetwork(self):
        return self.type_ in ["isolated", "management"]
    # end IsLibvirtNetwork

    def __str__(self):
        nw_str = f"Type : {self.type_}\n"
        nw_str += f"Bridge : {self.name_}\n"
        if self.type_ in ["nat", "management"]:
            if self.network4_ is not None:
                nw_str += "--- IPv4 Network ---\n"
                nw_str += f"Network : {self.network4_}\n"
                nw_str += f"Address : {self.ip4_}\n"
            if self.network6_ is not None:
                nw_str += "--- IPv6 Network ---\n"
                nw_str += f"Network : {self.network6_}\n"
                nw_str += f"Address : {self.ip6_}\n"
        nw_str += "===================================\n"
        return nw_str



import ipaddress
import logging
from socket import AddressFamily
import unittest
from unittest import mock

import deployer
import deployer.topology
import deployer.globals as G
from deployer.network import Network


class TestNetwork(unittest.TestCase):
    def setUp(self):
        G.OP = G.OP_CREATE
        self.nw_ = {"name": "test-nw", "type": "isolated"}
        logging.basicConfig(level="ERROR")

        # Mocking the psutil.net_if_addrs() method
        self.mocked_intf_ = [
            ('em1', AddressFamily.AF_INET, '20.20.1.1', '255.255.255.0', None, None),
            ('em1', AddressFamily.AF_INET6, '1234::1414:101',
                'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ff00', None, None),
            ('em1', AddressFamily.AF_PACKET, '52:54:00:aa:bb:cc', None, None, None)
        ]
        self.net_if_addr_patcher = mock.patch('psutil._psplatform.net_if_addrs')
        self.mock_net_if_addrs = self.net_if_addr_patcher.start()
        self.mock_net_if_addrs.return_value = self.mocked_intf_

        self.dummy_nw_ = Network({
            "name": "dummy-nw",
            "type": "nat",
            "subnet4": "11.11.11.0/24",
            "subnet6": "1234::11.11.11.0/120"
        })

        # Mocking the Topology.Networks() method
        self.mocked_topo_nw_ = {
            "dummy-nw": self.dummy_nw_
        }.items()
        self.topo_nw_patcher = mock.patch.object(deployer.topology.Topology, 'Networks')
        self.mock_topo_nw = self.topo_nw_patcher.start()
        self.mock_topo_nw.return_value = self.mocked_topo_nw_

    def tearDown(self):
        self.net_if_addr_patcher.stop()
        self.topo_nw_patcher.stop()

    def test_network_isolated_type(self):
        ok = Network._validate_network_conf(self.nw_)
        self.assertTrue(ok)

    def test_network_management_type(self):
        self.nw_["type"] = "management"
        self.nw_["subnet4"] = "10.10.1.0/24"

        ok = Network._has_valid_network_fields(self.nw_)
        self.assertTrue(ok)

    def test_network_nat_type(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"

        ok = Network._has_valid_network_fields(self.nw_)
        self.assertTrue(ok)

        self.nw_["subnet6"] = "1234::10.10.1.0/24"
        ok = Network._has_valid_network_fields(self.nw_)
        self.assertTrue(ok)

    def test_network_required_subnet4(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet6"] = "1234::10.10.1.0/24"

        ok = Network._has_valid_network_fields(self.nw_)
        self.assertFalse(ok)

        self.nw_["type"] = "management"
        ok = Network._has_valid_network_fields(self.nw_)
        self.assertFalse(ok)

    def test_validate_non_colliding_network_in_sys_v4(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"

        ok = Network._validate_network_conf(self.nw_)
        self.assertTrue(ok)

    def test_validate_colliding_network_in_sys_v4(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "20.20.1.0/24"

        ok = Network._validate_network_conf(self.nw_)
        self.assertFalse(ok)

    def test_validate_non_colliding_network_in_sys_v6(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"
        self.nw_["subnet6"] = "1234::10.10.1.0/120"

        ok = Network._validate_network_conf(self.nw_)
        self.assertTrue(ok)

    def test_validate_colliding_network_in_sys_v6(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"
        self.nw_["subnet6"] = "1234::20.20.1.0/120"

        ok = Network._validate_network_conf(self.nw_)
        self.assertFalse(ok)

    def test_validate_non_colliding_network_in_config_v4(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"

        ok = Network._validate_network_conf(self.nw_)
        self.assertTrue(ok)

    def test_validate_colliding_network_in_config_v4(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "11.11.11.0/24"

        ok = Network._validate_network_conf(self.nw_)
        self.assertFalse(ok)

    def test_validate_non_colliding_network_in_config_v6(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "10.10.1.0/24"
        self.nw_["subnet6"] = "1234::10.10.1.0/120"

        ok = Network._validate_network_conf(self.nw_)
        self.assertTrue(ok)

    def test_validate_colliding_network_in_config_v6(self):
        self.nw_["type"] = "nat"
        self.nw_["subnet4"] = "11.11.11.0/24"
        self.nw_["subnet6"] = "1234::11.11.11.0/120"

        ok = Network._validate_network_conf(self.nw_)
        self.assertFalse(ok)

    def test_add_v4_network(self):
        v4_nw_conf = {
            "name": "nw1",
            "type": "nat",
            "subnet4": "12.12.12.0/24"
        }
        v4_nw = Network(v4_nw_conf)
        self.assertEqual(v4_nw.network4_, ipaddress.IPv4Network('12.12.12.0/24'))
        self.assertEqual(v4_nw.ip4_, ipaddress.IPv4Network('12.12.12.0/24')[1])
        self.assertEqual(v4_nw.network6_, None)
        self.assertEqual(v4_nw.ip6_, None)

    def test_add_v6_network(self):
        nw_conf = {
            "name": "nw1",
            "type": "nat",
            "subnet4": "12.12.12.0/24",
            "subnet6": "1234::12.12.12.0/120"
        }
        nw1 = Network(nw_conf)
        self.assertEqual(nw1.network4_, ipaddress.IPv4Network('12.12.12.0/24'))
        self.assertEqual(nw1.ip4_, ipaddress.IPv4Network('12.12.12.0/24')[1])
        self.assertEqual(nw1.network6_, ipaddress.IPv6Network('1234::12.12.12.0/120'))
        self.assertEqual(nw1.ip6_, ipaddress.IPv6Network('1234::12.12.12.0/120')[1])



import sys
import logging
import yaml
import ipaddress
from pathlib import Path

import deployer.globals as G
from deployer.globals import LIBVIRT_IMAGES, VM_FLAVORS
from deployer.topology import Topology
from deployer.utils import ExecuteCommand


class VirtualMachine:
    name_ = ""
    vnc_port_ = -1
    user_data_cfg_ = ""
    network_data_cfg_ = ""
    root_disk_ = ""
    cloud_init_iso_ = ""
    networks_ = {}
    flavor_ = "pe"
    vcpus_ = 8
    ram_ = 16384
    root_disk_sz_ = "80G"
    base_image_ = ""
    topology_ = Topology()

    def __new__(cls, conf):
        if not VirtualMachine._validate_vm_config(conf):
            return None
        return super(VirtualMachine, cls).__new__(cls)
    # end __new__

    def __init__(self, conf):
        self.name_ = conf["name"]
        self.base_image_ = conf.get("base_image", G.OS_IMAGE_TEMPLATE)
        self.flavor_ = conf.get("flavor", "")
        self.vnc_port_ = int(conf["vnc_port"])
        self.libvirt_vm_base = LIBVIRT_IMAGES.joinpath(self.name_)
        self.user_data_cfg_ = self.libvirt_vm_base.joinpath("user-data.cfg")
        self.network_data_cfg_ = self.libvirt_vm_base.joinpath("nw-data.cfg")
        self.root_disk_ = self.libvirt_vm_base.joinpath("root_disk.qcow2")
        self.cloud_init_iso_ = self.libvirt_vm_base.joinpath("cloud-init.iso")

        self.libvirt_vm_base.mkdir(parents=True, exist_ok=True)
        self.networks_ = conf["networks"]

        if self.flavor_ == "ce":
            self.vcpus_ = 8
            self.ram_ = 8192
            self.root_disk_sz_ = "40G"
        elif self.flavor_ == "pe":
            self.vcpus_ = 8
            self.ram_ = 16384
            self.root_disk_sz_ = "80G"
        elif self.flavor_ == "dev":
            self.vcpus_ = 32
            self.ram_ = 65536
            self.root_disk_sz_ = "320G"

        if conf.get("vcpus", None) is not None:
            self.vcpus_ = int(conf["vcpus"])

        if conf.get("ram", None) is not None:
            self.ram_ = int(conf["ram"])

        if conf.get("disk", None) is not None:
            self.root_disk_sz_ = conf["disk"]
    # end __init__

    @staticmethod
    def _has_valid_vm_fields(vm):
        name = vm.get("name", None)
        if not name or not isinstance(name, str):
            logging.critical("'name' is required key in vm config")
            return False

        base_image = vm.get("base_image", None)
        if base_image:
            if not isinstance(base_image, str):
                logging.critical("'base_image' must be a string")
                return False
            if not Path(base_image).is_file():
                logging.critical(f"'base_image' file does not exist: {base_image}")
                return False

        networks = vm.get("networks", None)
        if not networks or not isinstance(networks, dict):
            logging.critical("'networks' is required key in vm config")
            return False

        vnc_port = vm.get("vnc_port", None)
        if not vnc_port or not isinstance(vnc_port, int) or vnc_port < 5900:
            logging.critical("'vnc_port' is required key in vm config")
            return False

        flavor = vm.get("flavor", None)
        vcpus = vm.get("vcpus", None)
        ram = vm.get("ram", None)
        disk = vm.get("disk", None)

        if not flavor:
            if not vcpus or not ram or not disk:
                logging.critical("'vcpu', 'ram' and 'disk' are mandatory if"
                                 " 'flavor' is not provided")
                return False
        else:
            if flavor not in VM_FLAVORS:
                logging.critical("Invalid VM flavor")
                return False

        if vcpus == 0 or ram == 0:
            logging.critical("'vcpu' or 'ram' is expected to be number >= 1")
            return False

        if disk == "":
            logging.critical("'disk' is expected to be non empty string")
            return False

        if vcpus and (not isinstance(vcpus, int) or vcpus < 1):
            logging.critical("'vcpu' is expected to be number >= 1")
            return False

        if ram and (not isinstance(ram, int) or ram < 1024):
            logging.critical("'ram' is expected to be number >= 1024")
            return False

        if disk and not isinstance(disk, str):
            logging.critical("'disk' is expected to be in str format")
            return False

        return True
    # end _has_valid_vm_fields

    @staticmethod
    def _validate_vm_network_config(nws):
        for name, ips in nws.items():
            network = VirtualMachine.Topology().GetNetwork(name)

            if network is None:
                logging.critical(f"Network {name} not found")
                return False

            if network.IsIsolated():
                return True

            try:
                ip4 = ipaddress.IPv4Address(ips["v4"])
                ip6 = None
                if ips.get("v6", None) is not None:
                    ip6 = ipaddress.ip_address(ips["v6"])
            except Exception as e:
                logging.error("Invalid Subnet(s). Please check again : ", e)
                return False

            # check if IP in IP Network
            if ip4 not in network.network4_:
                logging.error(f"{ip4} is not in {str(network.network4_)}")
                return False
            if ip6 and network.network6_:
                if ip6 not in network.network6_:
                    logging.error(f"{ip6} is not in {str(network.network6_)}")
                    return False

        return True
    # end _validate_vm_network_config

    @staticmethod
    def _validate_vm_config(vm):
        if G.NO_VM:
            return True

        if not VirtualMachine._has_valid_vm_fields(vm):
            return False

        if not VirtualMachine._validate_vm_network_config(vm["networks"]):
            return False

        return True
    # end _validate_vm_config

    @staticmethod
    def Topology():
        return VirtualMachine.topology_
    # end Topology

    def __generate_cloud_init_config(self):
        cloud_init = {
            "system_info": {
                "default_user": {
                    "name": G.BASE_OS,
                    "home": f"/home/{G.BASE_OS}"
                },
            },
            "password": G.BASE_OS,
            "chpasswd": {
                "expire": False
            },
            "hostname": f"{self.name_}-node",
            "ssh_pwauth": True
        }

        with open(self.user_data_cfg_, "w") as cloud_init_f:
            cloud_init_f.write("#cloud-config\n")
            yaml.dump(cloud_init, cloud_init_f, sort_keys=False)
            cloud_init_f.write("")
    # end __generate_cloud_init_config

    def __get_ubuntu_params(self):
        network_config = {
            "network": {
                "version": 2,
                "ethernets": {}
            }
        }
        interface_config = network_config["network"]["ethernets"]
        index = 1
        intf_fmt = "enp{}s0"
        return [network_config, interface_config, intf_fmt, index]
    # end __get_ubuntu_params

    def __get_rocky_params(self):
        network_config = {
            "version": 2,
            "ethernets": {}
        }
        interface_config = network_config["ethernets"]
        index = 0
        intf_fmt = "eth{}"
        return [network_config, interface_config, intf_fmt, index]
    # end __get_rocky_params

    def __get_os_based_params(self):
        if G.BASE_OS == "rocky":
            return self.__get_rocky_params()
        elif G.BASE_OS == "ubuntu":
            return self.__get_ubuntu_params()

        return [None, None, None, None]
    # end __get_os_based_params

    def __generate_netplan_config(self):
        network_config, interface_config, intf_fmt, index = self.__get_os_based_params()

        if ((network_config is None) or
                (interface_config is None) or
                (intf_fmt is None) or
                (index is None)):
            logging.critical("Error fetching os based network params")
            sys.exit(4)

        for nw in self.networks_:
            intf_name = intf_fmt.format(index)
            network = VirtualMachine.Topology().GetNetwork(nw)

            if network is None:
                logging.critical(f"Network {nw} not found")
                sys.exit(4)

            ip4 = self.networks_[nw].get("v4", None)
            ip6 = self.networks_[nw].get("v6", None)

            if network.IsManagement():
                interface_config[intf_name] = self.__generate_mgmt_config(ip4, network)
            elif network.IsNat():
                interface_config[intf_name] = self.__generate_nat_config(ip4, ip6, network)
            elif network.IsIsolated():
                interface_config[intf_name] = self.__generate_iso_config(ip4, ip6)
            index = index + 1

        with open(self.network_data_cfg_, "w") as nw_config_f:
            nw_config_f.write("#network-config\n")
            yaml.dump(network_config, nw_config_f, sort_keys=False)
    # end __generate_netplan_config

    def __generate_mgmt_config(self, ip4, nw):
        addr = ""
        if nw.network4_ is not None:
            addr = "{}/{}".format(ip4, nw.network4_.prefixlen)
        mgmt_nw_config = {
            "addresses": [addr],
            "routes": [{
                "to": str(nw.network4_),
                "via": str(nw.ip4_)
            }]
        }
        return mgmt_nw_config
    # end __generate_mgmt_config

    def __generate_nat_config(self, ip4, ip6, nw):
        addresses = []

        if nw.network4_ is not None:
            addresses.append(f"{ip4}/{nw.network4_.prefixlen}") if ip4 is not None else None

        if nw.network6_ is not None:
            addresses.append(f"{ip6}/{nw.network6_.prefixlen}") if ip6 is not None else None

        nat_nw_config = {
            "addresses": addresses,
            "nameservers": {
                "addresses": ['66.129.233.81', '10.84.5.101', '172.21.200.60'],
                "search": ['jnpr.net', 'juniper.net', 'englab.juniper.net']
            },
            "gateway4": str(nw.ip4_),
        }

        if nw.network6_ is not None:
            nat_nw_config["gateway6"] = str(nw.ip6_)

        return nat_nw_config
    # end __generate_nat_config

    def __generate_iso_config(self, ip4, ip6):
        addresses = []
        routes = []

        if ip4 is not None:
            addresses.append(f"{ip4}/24")
            nw = ipaddress.IPv4Network(f"{ip4}/24", strict=False)
            routes.append({
                "to": str(nw),
                "scope": "link"
            })

        if ip6 is not None:
            addresses.append(f"{ip6}/120")
            nw = ipaddress.IPv6Network(f"{ip6}/120", strict=False)
            routes.append({
                "to": str(nw),
                "scope": "link"
            })

        iso_nw_config = {
            "addresses": addresses,
            "routes": routes
        }
        return iso_nw_config
    # end __generate_iso_config

    def __create_root_disk(self):
        cmd = f"sudo qemu-img convert -f qcow2 -O qcow2 {self.base_image_} {self.root_disk_}"
        ExecuteCommand(cmd)

        cmd = f"sudo qemu-img resize {self.root_disk_} {self.root_disk_sz_}"
        ExecuteCommand(cmd)
    # end __create_root_disk

    def __generate_cloud_init_iso(self):
        cmd = f"sudo cloud-localds -N {self.network_data_cfg_} "\
                f"{self.cloud_init_iso_} {self.user_data_cfg_}"
        ExecuteCommand(cmd)
    # end __generate_cloud_init_iso

    def __execute_virt_install_cmd(self):
        name_arg = f"--name {self.name_}"
        graphic_arg = f"--graphics vnc,listen=0.0.0.0,port={self.vnc_port_}"
        root_disk_arg = f"--disk path={self.root_disk_},bus=virtio,format=qcow2"
        cloud_init_arg = f"--disk {self.cloud_init_iso_},device=cdrom"
        cpu_ram_arg = f"--vcpus {self.vcpus_} --ram {self.ram_}"

        network_args = ""
        for nw in self.networks_:
            tmp_nw_cfg = f"--network bridge={nw},model=virtio "
            network_args += tmp_nw_cfg

        os_var = ""
        if G.BASE_OS == "ubuntu":
            os_var = "ubuntu20.04"
        elif G.BASE_OS == "rocky":
            os_var = "rhel8.0"

        cmd = f"sudo virt-install --virt-type kvm --os-variant {os_var} " \
              f"{name_arg} {cpu_ram_arg} {graphic_arg} " \
              f"{root_disk_arg} {cloud_init_arg} {network_args} " \
              "--noautoconsole --import"
        ExecuteCommand(cmd)
    # end __execute_virt_install_cmd

    def Create(self):
        logging.info(f"Creating Virtual Machine : {self.name_}")

        self.__generate_cloud_init_config()
        self.__generate_netplan_config()
        self.__create_root_disk()
        self.__generate_cloud_init_iso()

        self.__execute_virt_install_cmd()
    # end Create

    def Delete(self):
        logging.info(f"Deleting Virtual Machine : {self.name_}")

        cmd = f"sudo virsh destroy {self.name_}"
        ExecuteCommand(cmd)

        cmd = f"sudo virsh undefine --remove-all-storage {self.name_}"
        ExecuteCommand(cmd)
    # end Delete

    def __str__(self):
        vm_str = ""
        vm_str += f"Flavor : {self.flavor_}\n"
        vm_str += f"VNC : {self.vnc_port_}\n"
        vm_str += "Networks : \n"
        for nw, ips in self.networks_.items():
            ip4 = ips.get("v4", None)
            ip6 = ips.get("v6", None)
            vm_str += f"\t{nw} :\n\t\tv4 : {ip4}\n\t\tv6 : {ip6}\n"
        vm_str += "===================================\n"
        return vm_str
    # end __str__



import logging
import unittest

from deployer.deployer import Deployer


class TestDeployer(unittest.TestCase):
    def setUp(self):
        self.config_ = {"version": 2, "networks": [], "vms": []}
        self.deployer_ = Deployer(self.config_)
        logging.basicConfig(level='CRITICAL')

    def tearDown(self):
        del self.deployer_

    def test_valid_version(self):
        ok = self.deployer_._is_valid_version()
        self.assertTrue(ok)

    def test_invalid_version(self):
        self.deployer_.config_ = {"version": 3, "networks": [], "vms": []}
        ok = self.deployer_._is_valid_version()
        self.assertFalse(ok)

    def test_missing_version(self):
        self.deployer_.config_ = {"networks": [], "vms": []}
        ok = self.deployer_._is_valid_version()
        self.assertFalse(ok)

    def test_parse_empty_networks(self):
        ok = self.deployer_._parse_networks()
        self.assertFalse(ok)

    def test_parse_non_list_networks(self):
        self.deployer_.config_['networks'] = {}
        ok = self.deployer_._parse_networks()
        self.assertFalse(ok)

    def test_parse_missing_networks(self):
        del self.deployer_.config_['networks']
        ok = self.deployer_._parse_networks()
        self.assertFalse(ok)

    def test_parse_vms(self):
        ok = self.deployer_._parse_vms()
        self.assertFalse(ok)

    def test_parse_non_list_vms(self):
        self.deployer_.config_['vms'] = {}
        ok = self.deployer_._parse_vms()
        self.assertFalse(ok)

    def test_parse_missing_vms(self):
        del self.deployer_.config_['vms']
        ok = self.deployer_._parse_vms()
        self.assertFalse(ok)



import setuptools

with open("README.md", "r", encoding="utf-8") as fhand:
    long_description = fhand.read()

setuptools.setup(
    name="topology-deployer",
    version="0.2.1",
    author="Ved Patel",
    description=("A python tool to deploy topology "
                "defined in a json config"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vpatel95/libvirt-topology",
    project_urls={
        "Bug Tracker": "https://github.com/vpatel95/libvirt-topology/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=["pyyaml", "ipaddress", "psutil", "pytest"],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "topology-deployer = deployer.deployer:topology_deployer",
        ]
    }
)



import logging
import unittest
from unittest import mock

import deployer
import deployer.topology
import deployer.globals as G
from deployer.deployer import Deployer
from deployer.network import Network
from deployer.virtual_machine import VirtualMachine


class TestVirtualMachine(unittest.TestCase):
    def setUp(self):
        self.deployer_ = Deployer({})
        self.vm_ = None
        self.iso_nw_ = Network({
            "name": "iso1",
            "type": "isolated"
        })
        self.mgmt_nw_ = Network({
            "name": "mgmt1",
            "type": "management",
            "subnet4": "10.1.1.0/24",
        })
        self.nat_nw_ = Network({
            "name": "nat1",
            "type": "nat",
            "subnet4": "20.1.1.0/24",
            "subnet6": "1234::20.1.1.0/120",
        })
        self.mgmt_conf_ = {"v4": "10.1.1.1"}
        self.nat_conf_ = {"v4": "20.1.1.1", "v6": "1234::20.1.1.1"}
        self.iso_conf_ = {"v4": "30.1.1.1", "v6": "1234::30.1.1.1"}
        self.vm_conf_ = {
            "name": "vm1",
            "networks": {
                "mgmt1": self.mgmt_conf_,
                "nat1": self.nat_conf_,
                "iso1": self.iso_conf_
            },
            "vnc_port": 5900,
            "flavor": "pe",
            "vcpus": 8,
            "ram": 16384,
            "disk": "80G"
        }
        logging.basicConfig(level="INFO")

        # Mocking os.mkdir() method
        self.mkdir_patcher = mock.patch('pathlib.PosixPath.mkdir')
        self.mock_mkdir = self.mkdir_patcher.start()

        # Mocking the Topology.GetNetwork() method
        def get_network(name):
            if name == "mgmt1":
                return self.mgmt_nw_
            elif name == "nat1":
                return self.nat_nw_
            elif name == "iso1":
                return self.iso_nw_
            else:
                return None
        # end get_network

        self.topo_getnw_patcher = mock.patch.object(deployer.topology.Topology, 'GetNetwork')
        self.mock_topo_getnw = self.topo_getnw_patcher.start()
        self.mock_topo_getnw.side_effect = get_network

    def tearDown(self):
        self.topo_getnw_patcher.stop()
        self.mock_mkdir = self.mkdir_patcher.stop()

    def test_name_validity(self):
        # Test empty name. It should fail
        self.vm_conf_["name"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test non string name. It should fail
        self.vm_conf_["name"] = {}
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        self.vm_conf_["name"] = 10
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid name. It should pass
        self.vm_conf_["name"] = "vm1"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_networks_validity(self):
        # Test empty networks. It should fail
        self.vm_conf_["networks"] = {}
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test non dict networks. It should fail
        self.vm_conf_["networks"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid networks. It should pass
        self.vm_conf_["networks"] = {"mgmt": {"v4": ""}}
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_vnc_port_validity(self):
        # Test empty vnc_port. It should fail
        self.vm_conf_["vnc_port"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test non int vnc_port. It should fail
        self.vm_conf_["vnc_port"] = "5900"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test vnc_port < 5900. It should fail
        self.vm_conf_["vnc_port"] = 5899
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid vnc_port. It should pass
        self.vm_conf_["vnc_port"] = 5900
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_flavor_validity(self):
        # Test empty flavor. It should pass as vcpus, ram and disk are not empty
        self.vm_conf_["flavor"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Delete vcpus from the dict. Empty flavor should fail
        del self.vm_conf_["vcpus"]
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)
        self.vm_conf_["vcpus"] = 8

        # Test non string flavor. It should fail
        self.vm_conf_["flavor"] = 10
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test flavor not in VM_FLAVORS. It should fail
        self.vm_conf_["flavor"] = "test"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid flavor. It should pass
        self.vm_conf_["flavor"] = "pe"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_vcpus_validity(self):
        # Empty vcpu with valid flavor. It should pass
        del self.vm_conf_["vcpus"]
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Test empty vcpus. It should pass
        self.vm_conf_["vcpus"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Test non int vcpus. It should fail
        self.vm_conf_["vcpus"] = "8"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test vcpus < 1. It should fail
        self.vm_conf_["vcpus"] = 0
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid vcpus. It should pass
        self.vm_conf_["vcpus"] = 8
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_ram_validity(self):
        # Empty ram with valid flavor. It should pass
        del self.vm_conf_["ram"]
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Test empty ram. It should pass
        self.vm_conf_["ram"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Test non int ram. It should fail
        self.vm_conf_["ram"] = "16384"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test ram < 1. It should fail
        self.vm_conf_["ram"] = 0
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid ram. It should pass
        self.vm_conf_["ram"] = 16384
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_disk_validity(self):
        # Empty disk with valid flavor. It should pass
        del self.vm_conf_["disk"]
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

        # Test empty disk. It should pass
        self.vm_conf_["disk"] = ""
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test non string disk. It should fail
        self.vm_conf_["disk"] = 80
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertFalse(ok)

        # Test valid disk. It should pass
        self.vm_conf_["disk"] = "80G"
        ok = VirtualMachine._has_valid_vm_fields(self.vm_conf_)
        self.assertTrue(ok)

    def test_add_pe_virtual_machine(self):
        self.vm_ = VirtualMachine(self.vm_conf_)
        self.mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        self.assertEqual(self.vm_.name_, "vm1")
        self.assertEqual(self.vm_.vnc_port_, 5900)
        self.assertEqual(self.vm_.flavor_, "pe")
        self.assertEqual(self.vm_.vcpus_, 8)
        self.assertEqual(self.vm_.ram_, 16384)
        self.assertEqual(self.vm_.root_disk_sz_, "80G")
        self.assertEqual(self.vm_.networks_, {
            "mgmt1": self.mgmt_conf_,
            "nat1": self.nat_conf_,
            "iso1": self.iso_conf_
        })
        self.assertEqual(str(self.vm_.user_data_cfg_),
                         "/var/lib/libvirt/images/vm1/user-data.cfg")
        self.assertEqual(str(self.vm_.network_data_cfg_),
                         "/var/lib/libvirt/images/vm1/nw-data.cfg")
        self.assertEqual(str(self.vm_.root_disk_),
                         "/var/lib/libvirt/images/vm1/root_disk.qcow2")
        self.assertEqual(str(self.vm_.cloud_init_iso_),
                         "/var/lib/libvirt/images/vm1/cloud-init.iso")
        self.assertEqual(str(self.vm_.libvirt_vm_base),
                         "/var/lib/libvirt/images/vm1")

    def test_add_ce_virtual_machine(self):
        self.vm_conf_["flavor"] = "ce"
        del self.vm_conf_["disk"]
        del self.vm_conf_["ram"]
        del self.vm_conf_["vcpus"]
        self.vm_ = VirtualMachine(self.vm_conf_)
        self.mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        self.assertEqual(self.vm_.name_, "vm1")
        self.assertEqual(self.vm_.vnc_port_, 5900)
        self.assertEqual(self.vm_.flavor_, "ce")
        self.assertEqual(self.vm_.vcpus_, 8)
        self.assertEqual(self.vm_.ram_, 8192)
        self.assertEqual(self.vm_.root_disk_sz_, "40G")
        self.assertEqual(self.vm_.networks_, {
            "mgmt1": self.mgmt_conf_,
            "nat1": self.nat_conf_,
            "iso1": self.iso_conf_
        })
        self.assertEqual(str(self.vm_.user_data_cfg_),
                         "/var/lib/libvirt/images/vm1/user-data.cfg")
        self.assertEqual(str(self.vm_.network_data_cfg_),
                         "/var/lib/libvirt/images/vm1/nw-data.cfg")
        self.assertEqual(str(self.vm_.root_disk_),
                         "/var/lib/libvirt/images/vm1/root_disk.qcow2")
        self.assertEqual(str(self.vm_.cloud_init_iso_),
                         "/var/lib/libvirt/images/vm1/cloud-init.iso")
        self.assertEqual(str(self.vm_.libvirt_vm_base),
                         "/var/lib/libvirt/images/vm1")

    def test_add_custom_virtual_machine(self):
        del self.vm_conf_["flavor"]
        self.vm_conf_["disk"] = "300G"
        self.vm_conf_["ram"] = 24576
        self.vm_conf_["vcpus"] = 12
        self.vm_ = VirtualMachine(self.vm_conf_)
        self.mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        self.assertEqual(self.vm_.name_, "vm1")
        self.assertEqual(self.vm_.vnc_port_, 5900)
        self.assertEqual(self.vm_.flavor_, "")
        self.assertEqual(self.vm_.vcpus_, 12)
        self.assertEqual(self.vm_.ram_, 24576)
        self.assertEqual(self.vm_.root_disk_sz_, "300G")
        self.assertEqual(self.vm_.networks_, {
            "mgmt1": self.mgmt_conf_,
            "nat1": self.nat_conf_,
            "iso1": self.iso_conf_
        })
        self.assertEqual(str(self.vm_.user_data_cfg_),
                         "/var/lib/libvirt/images/vm1/user-data.cfg")
        self.assertEqual(str(self.vm_.network_data_cfg_),
                         "/var/lib/libvirt/images/vm1/nw-data.cfg")
        self.assertEqual(str(self.vm_.root_disk_),
                         "/var/lib/libvirt/images/vm1/root_disk.qcow2")
        self.assertEqual(str(self.vm_.cloud_init_iso_),
                         "/var/lib/libvirt/images/vm1/cloud-init.iso")
        self.assertEqual(str(self.vm_.libvirt_vm_base),
                         "/var/lib/libvirt/images/vm1")

    @mock.patch('pathlib.Path.is_file', return_value=True)
    def test_add_virtual_machine_valid_base_image(self, mock_is_file):
        self.vm_conf_["base_image"] = "/path/to/base_image"
        self.deployer_.config_['vms'] = [self.vm_conf_]
        ok = self.deployer_._parse_vms()
        self.vm_ = self.deployer_.topology_.vms_[self.vm_conf_["name"]]
        self.assertEqual(self.vm_.base_image_, "/path/to/base_image")
        self.assertTrue(ok)

    @mock.patch('pathlib.Path.is_file', return_value=False)
    def test_add_virtual_machine_invalid_base_image(self, mock_is_file):
        self.vm_conf_["base_image"] = "/path/to/base_image"
        self.deployer_.config_['vms'] = [self.vm_conf_]
        ok = self.deployer_._parse_vms()
        self.vm_ = self.deployer_.topology_.vms_[self.vm_conf_["name"]]
        self.assertFalse(ok)
        del self.deployer_.topology_.vms_[self.vm_conf_["name"]]

    def test_add_virtual_machine_default_base_image(self):
        self.deployer_.config_['vms'] = [self.vm_conf_]
        ok = self.deployer_._parse_vms()
        self.vm_ = self.deployer_.topology_.vms_[self.vm_conf_["name"]]
        self.assertEqual(self.vm_.base_image_, G.UBUNTU_TEMPLATE)
        self.assertTrue(ok)
